package com.lanren.huhu.partner.manager;

import com.lanren.huhu.partner.domain.PartnerAccount;
import com.lanren.huhu.partner.domain.UserAgent;
import com.lanren.huhu.partner.model.ParentAgent;
import com.lanren.huhu.partner.model.ParentPartner;
import com.lanren.huhu.partner.model.User;
import com.lanren.huhu.partner.service.PartnerAccountService;
import com.lanren.huhu.partner.service.UserAgentService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheConfig;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.stereotype.Component;

import java.util.ArrayList;

/**
 * @author houseme
 */
@Component
@CacheConfig(cacheNames = "user_cache")
public class UserManager {
    private static Logger logger = LoggerFactory.getLogger(UserManager.class);

    @Autowired
    private PartnerAccountService partnerAccountService;

    @Autowired
    private UserAgentService userAgentService;

    /**
     * @description: 根据用户id, 查找一个用户的全部合伙人上级 和 代理商上级
     *       缓存用户邀请关系和代理商关系数据
     *       缓存不能带有构造方法的Bean 所以只能返回User的Bean
     *       包装Result在Cacheable注解的方法之外进行操作
     * @param userId
     * @return User
     * @throws
     * @author chen
     * @date 2019-06-26 20:01
     */
    @Cacheable
    public User getRelationByUserId(Integer userId) {
        logger.info("new a user object ");
        if (logger.isDebugEnabled()) {
            logger.info("new a user object ");
        }
        User user = new User();
        user.setUserId(userId);
        setPartnerList(user);
        setAgentList(user);
        return user;
    }

    /**
     * 清理用户缓存
     */
    @CacheEvict(key = "#userId")
    public void deleteRalationByUserId(Integer userId) {
        if (logger.isInfoEnabled()) {
            logger.info("delete user relation cache, userId: {}", userId);
        }
    }

    /**
     * @description: 查找一个用户的全部合伙人上级, set到User对象的partnerList中
     * @param user
     * @author chen
     * @date 2019-06-26 20:37
     */
    private void setPartnerList(User user) {
        if (logger.isInfoEnabled()) {
            logger.info("begin setPartnerList for user: {} ......", user.getUserId());
        }
        if (null == user.getPartnerList()) {
            ArrayList<ParentPartner> partnerList = new ArrayList<ParentPartner>();
            user.setPartnerList(partnerList);
        }
        /**
         * 无限找
         */
        int loopUserId = user.getUserId();
        while (true) {
            if (logger.isDebugEnabled()) {
                logger.info("now loop uid is {}", loopUserId);
            }
            ParentPartner parentPartner = getParentPartner(loopUserId);
            if (null == parentPartner) {
                break;
            } else {
                if (logger.isDebugEnabled()) {
                    logger.info("parent is {} ", parentPartner.getUserId());
                }
                user.getPartnerList().add(parentPartner);
                loopUserId = parentPartner.getUserId();
            }
        }
        if (logger.isInfoEnabled()) {
            logger.info("setPartnerList for user: {} end", user.getUserId());
        }
    }

    /**
     * @description: 查找一个用户的全部代理商上级 set到User对象的agentList中
     * @param user
     * @date 2019-06-27 11:16
     */
    private void setAgentList(User user) {
        if (logger.isInfoEnabled()) {
            logger.info("begin setAgentList for user: {} ......", user.getUserId());
        }
        if (null == user.getAgentList()) {
            ArrayList<ParentAgent> agentList = new ArrayList<ParentAgent>();
            user.setAgentList(agentList);
        }
        /**
         * 无限找
         * 先无限找上级合伙人, 直到找到第一个是代理商的合伙人
         */
        int loopUserId = user.getUserId();
        while (true) {
            if (logger.isDebugEnabled()) {
                logger.info("now loop uid is {}", loopUserId);
            }
            ParentPartner parentPartner = getParentPartner(loopUserId);
            /**
             * 有上级合伙人
             */
            if (null != parentPartner) {
                UserAgent userAgent = userAgentService.getOneByUserId(parentPartner.getUserId());
                /**
                 * 找到了第一个是代理商的合伙人
                 */
                if (null != userAgent) {
                    /**
                     * 然后开始 无限找 这个人的上级代理商
                     */
                    if (logger.isDebugEnabled()) {
                        logger.info("find first agent({}) for user: {}", userAgent.getAgentId(), user.getUserId());
                    }
                    user.getAgentList().add(new ParentAgent(userAgent.getUserId(), userAgent.getAgentId(), userAgent.getAgentLevel()));
                    if (logger.isDebugEnabled()) {
                        logger.info("loop find parent agent start ......");
                    }
                    int loogAgentId = userAgent.getAgentId();
                    while (true) {
                        ParentAgent loopUserAgent = getParentAgent(loogAgentId);
                        if (null == loopUserAgent) {
                            break;
                        } else {
                            if (logger.isDebugEnabled()) {
                                logger.info("parent agent is ", loopUserAgent.getAgentId());
                            }
                            user.getAgentList().add(new ParentAgent(loopUserAgent.getUserId(), loopUserAgent.getAgentId(), loopUserAgent.getLevel()));
                            loogAgentId = loopUserAgent.getAgentId();
                        }
                    }
                    /**
                     * 找完 "第一个是代理商的合伙人" 的 全部上级代理后 退出while循环
                     */
                    if (logger.isDebugEnabled()) {
                        logger.info("loop find parent agent end");
                    }
                    break;
                } else {
                    /**
                     * 这个人不是代理商 继续往上找
                     */
                    loopUserId = parentPartner.getUserId();
                }
            } else {
                /**
                 * 没有上级合伙人 打印结束信息 跳出循环
                 */
                if (logger.isDebugEnabled()) {
                    logger.info("loop end with userId: {}", loopUserId);
                }
                break;
            }
        }
        if (logger.isInfoEnabled()) {
            logger.info("setAgentList for user: {} end", user.getUserId());
        }
    }

    /**
     * 按userId查找上级合伙人的 userId 和 level
     * @param userId
     * @return
     */
    private ParentPartner getParentPartner(Integer userId) {
        PartnerAccount partnerAccount = partnerAccountService.getParentPartner(userId);
        ParentPartner parentPartner = null;
        if (null != partnerAccount) {
            int partnerLevel = partnerAccount.getIsSuperPartner() == 1 ? 30 : partnerAccount.getPartnerLevel();
            parentPartner = new ParentPartner(partnerAccount.getUserId(),partnerLevel);
        }
        return parentPartner;
    }

    /**
     * 按agentId 查找上级代理商的 agentId 和 level
     * @param agentId
     * @return
     */
    private ParentAgent getParentAgent(Integer agentId) {
        UserAgent userAgent = userAgentService.getParentAgent(agentId);
        ParentAgent parentAgent = null;
        if (null != userAgent) {
            int agentLevel = userAgent.getAgentLevel();
            parentAgent = new ParentAgent(userAgent.getUserId(), userAgent.getAgentId(), agentLevel);
        }
        return parentAgent;
    }

    /**
     * 根据输入的用户id, 刷新这个用户以及他邀请关系的所有下级的邀请关系
     * @param userId
     * @return
     *
     * 废弃掉这个方法, 这个方法中new出的util对象, 会使用UserManager.this调用加上 清缓存 和 重写缓存 注解的方法
     * 这样是不走缓存的, 也清不掉缓存的
     *
     * 改为从上级调用 从controller中
     */
//    public boolean changeRelation(int userId) {
//        PartnerTreeUtil partnerTreeUtil = new PartnerTreeUtil();
//        return partnerTreeUtil.findPartnerChiledren(userId, partnerAccountService, this);
//    }
}
